@@ -3,6 +3,7 @@
  * gdl-dock-item.c
  *
  * Author: Gustavo Girldez <gustavo.giraldez@gmx.net>
+ *         Naba Kumar  <naba@gnome.org>
  *
  * Based on GnomeDockItem/BonoboDockItem.  Original copyright notice follows.
  *
@@ -48,6 +49,7 @@
 #include "libgdltypebuiltins.h"
 #include "libgdlmarshal.h"
 
+#define NEW_DOCK_ITEM_RATIO 0.3
 
 /* ----- Private prototypes ----- */
 
@@ -443,7 +445,7 @@ gdl_dock_item_set_property  (GObject      *g_object,
                 item->behavior &= ~GDL_DOCK_ITEM_BEH_LOCKED;
 
             if (old_beh ^ item->behavior) {
-		gdl_dock_item_showhide_grip (item);
+                gdl_dock_item_showhide_grip (item);
                 g_object_notify (g_object, "behavior");
 
                 if (GDL_DOCK_OBJECT_GET_MASTER (item))
@@ -689,6 +691,10 @@ gdl_dock_item_size_allocate (GtkWidget     *widget,
 
     widget->allocation = *allocation;
 
+    /* Once size is allocated, preferred size is no longer necessary */
+    item->_priv->preferred_height = -1;
+    item->_priv->preferred_width = -1;
+    
     if (GTK_WIDGET_REALIZED (widget))
         gdk_window_move_resize (widget->window,
                                 widget->allocation.x,
@@ -1145,35 +1151,118 @@ gdl_dock_item_dock (GdlDockObject    *object,
     guint	   available_space=0;
     gint	   pref_size=-1;
     guint	   splitpos=0;
-    GtkRequisition req;
-  	
+    GtkRequisition req, object_req, parent_req;
+    
     parent = gdl_dock_object_get_parent_object (object);
     gdl_dock_item_preferred_size (GDL_DOCK_ITEM (requestor), &req);
-
+    gdl_dock_item_preferred_size (GDL_DOCK_ITEM (object), &object_req);
+    if (GDL_IS_DOCK_ITEM (parent))
+        gdl_dock_item_preferred_size (GDL_DOCK_ITEM (parent), &parent_req);
+    else
+    {
+        parent_req.height = GTK_WIDGET (parent)->allocation.height;
+        parent_req.width = GTK_WIDGET (parent)->allocation.width;
+    }
+    
+    /* If preferred size is not set on the requestor (perhaps a new item),
+     * then estimate and set it. The default value (either 0 or 1 pixels) is
+     * not any good.
+     */
+    switch (position) {
+        case GDL_DOCK_TOP:
+        case GDL_DOCK_BOTTOM:
+            if (req.width < 2)
+            {
+                req.width = object_req.width;
+                g_object_set (requestor, "preferred-width", req.width, NULL);
+            }
+            if (req.height < 2)
+            {
+                req.height = NEW_DOCK_ITEM_RATIO * object_req.height;
+                g_object_set (requestor, "preferred-height", req.height, NULL);
+            }
+            if (req.width > 1)
+                g_object_set (object, "preferred-width", req.width, NULL);
+            if (req.height > 1)
+                g_object_set (object, "preferred-height",
+                              object_req.height - req.height, NULL);
+            break;
+        case GDL_DOCK_LEFT:
+        case GDL_DOCK_RIGHT:
+            if (req.height < 2)
+            {
+                req.height = object_req.height;
+                g_object_set (requestor, "preferred-height", req.height, NULL);
+            }
+            if (req.width < 2)
+            {
+                req.width = NEW_DOCK_ITEM_RATIO * object_req.width;
+                g_object_set (requestor, "preferred-width", req.width, NULL);
+            }
+            if (req.height > 1)
+                g_object_set (object, "preferred-height", req.height, NULL);
+            if (req.width > 1)
+                g_object_set (object, "preferred-width",
+                          object_req.width - req.width, NULL);
+            break;
+        case GDL_DOCK_CENTER:
+            if (req.height < 2)
+            {
+                req.height = object_req.height;
+                g_object_set (requestor, "preferred-height", req.height, NULL);
+            }
+            if (req.width < 2)
+            {
+                req.width = object_req.width;
+                g_object_set (requestor, "preferred-width", req.width, NULL);
+            }
+            if (req.height > 1)
+                g_object_set (object, "preferred-height", req.height, NULL);
+            if (req.width > 1)
+                g_object_set (object, "preferred-width", req.width, NULL);
+            break;
+        default: 
+        {
+            GEnumClass *enum_class = G_ENUM_CLASS (g_type_class_ref (GDL_TYPE_DOCK_PLACEMENT));
+            GEnumValue *enum_value = g_enum_get_value (enum_class, position);
+            const gchar *name = enum_value ? enum_value->value_name : NULL;
+            
+            g_warning (_("Unsupported docking strategy %s in dock object of type %s"),
+                       name,  G_OBJECT_TYPE_NAME (object));
+            g_type_class_unref (enum_class);
+            return;
+        }
+    }
     switch (position) {
         case GDL_DOCK_TOP:
         case GDL_DOCK_BOTTOM:
             /* get a paned style dock object */
             new_parent = g_object_new (gdl_dock_object_type_from_nick ("paned"),
                                        "orientation", GTK_ORIENTATION_VERTICAL,
+                                       "preferred-width", object_req.width,
+                                       "preferred-height", object_req.height,
                                        NULL);
             add_ourselves_first = (position == GDL_DOCK_BOTTOM);
             if (parent)
-            	available_space=GTK_WIDGET(parent)->allocation.height;
+                available_space = parent_req.height;
             pref_size = req.height;
             break;
         case GDL_DOCK_LEFT:
         case GDL_DOCK_RIGHT:
             new_parent = g_object_new (gdl_dock_object_type_from_nick ("paned"),
                                        "orientation", GTK_ORIENTATION_HORIZONTAL,
+                                       "preferred-width", object_req.width,
+                                       "preferred-height", object_req.height,
                                        NULL);
             add_ourselves_first = (position == GDL_DOCK_RIGHT);
             if(parent)
-            	available_space = GTK_WIDGET(parent)->allocation.width;
+                available_space = parent_req.width;
             pref_size = req.width;
             break;
         case GDL_DOCK_CENTER:
             new_parent = g_object_new (gdl_dock_object_type_from_nick ("notebook"),
+                                       "preferred-width", object_req.width,
+                                       "preferred-height", object_req.height,
                                        NULL);
             add_ourselves_first = TRUE;
             break;
@@ -1181,7 +1270,7 @@ gdl_dock_item_dock (GdlDockObject    *object,
         {
             GEnumClass *enum_class = G_ENUM_CLASS (g_type_class_ref (GDL_TYPE_DOCK_PLACEMENT));
             GEnumValue *enum_value = g_enum_get_value (enum_class, position);
-            gchar *name = enum_value ? enum_value->value_name : NULL;
+            const gchar *name = enum_value ? enum_value->value_name : NULL;
             
             g_warning (_("Unsupported docking strategy %s in dock object of type %s"),
                        name,  G_OBJECT_TYPE_NAME (object));
@@ -1210,7 +1299,7 @@ gdl_dock_item_dock (GdlDockObject    *object,
     if (add_ourselves_first) {
         gtk_container_add (GTK_CONTAINER (new_parent), GTK_WIDGET (object));
         gtk_container_add (GTK_CONTAINER (new_parent), GTK_WIDGET (requestor));
-	splitpos = available_space - pref_size;
+        splitpos = available_space - pref_size;
     } else {
         gtk_container_add (GTK_CONTAINER (new_parent), GTK_WIDGET (requestor));
         gtk_container_add (GTK_CONTAINER (new_parent), GTK_WIDGET (object));
@@ -1232,8 +1321,8 @@ gdl_dock_item_dock (GdlDockObject    *object,
         g_object_set (G_OBJECT (new_parent),
                       "position", g_value_get_uint (other_data),
                       NULL);
-    }else if (splitpos > 0 && splitpos < available_space) {
-			g_object_set (G_OBJECT (new_parent), "position", splitpos, NULL);
+    } else if (splitpos > 0 && splitpos < available_space) {
+        g_object_set (G_OBJECT (new_parent), "position", splitpos, NULL);
     }
     
     GDL_DOCK_OBJECT_UNSET_FLAGS (object, GDL_DOCK_IN_REFLOW);
@@ -1268,25 +1357,25 @@ gdl_dock_item_popup_menu (GdlDockItem  *item,
                                    GTK_WIDGET (item),
                                    gdl_dock_item_detach_menu);
         
-	if (item->behavior & GDL_DOCK_ITEM_BEH_LOCKED) {
+        if (item->behavior & GDL_DOCK_ITEM_BEH_LOCKED) {
             /* UnLock menuitem */
             mitem = gtk_menu_item_new_with_label (_("UnLock"));
             gtk_menu_shell_append (GTK_MENU_SHELL (item->_priv->menu), 
-			           mitem);
+                                   mitem);
             g_signal_connect (mitem, "activate",
                               G_CALLBACK (gdl_dock_item_unlock_cb), item);
-	} else {
+        } else {
             /* Hide menuitem. */
             mitem = gtk_menu_item_new_with_label (_("Hide"));
             gtk_menu_shell_append (GTK_MENU_SHELL (item->_priv->menu), mitem);
             g_signal_connect (mitem, "activate", 
                               G_CALLBACK (gdl_dock_item_hide_cb), item);
-	    /* Lock menuitem */
+            /* Lock menuitem */
             mitem = gtk_menu_item_new_with_label (_("Lock"));
             gtk_menu_shell_append (GTK_MENU_SHELL (item->_priv->menu), mitem);
             g_signal_connect (mitem, "activate",
                               G_CALLBACK (gdl_dock_item_lock_cb), item);
-	}
+        }
     }
 
     /* Show popup menu. */
@@ -1628,36 +1717,35 @@ gdl_dock_item_hide_item (GdlDockItem *item)
        restore the position later */
     if (!GDL_DOCK_OBJECT_AUTOMATIC (item)) {
         if (item->_priv->ph)
-			g_object_unref (item->_priv->ph); 
+            g_object_unref (item->_priv->ph); 
         
-		gboolean isFloating = FALSE;
-		gint width=0, height=0, x=0, y = 0;
-		
-		
-		if( GDL_IS_DOCK (gdl_dock_object_get_parent_object (GDL_DOCK_OBJECT (item))))
-		{
-			GdlDock* dock = GDL_DOCK (gdl_dock_object_get_parent_object (GDL_DOCK_OBJECT (item)));
-			g_object_get (dock,
-						  "floating", &isFloating, 
-						  "width", &width,
-						  "height",&height,
-						  "floatx",&x,
-						  "floaty",&y,
-						  NULL);
-		}else{
-			item->_priv->preferred_width=GTK_WIDGET (item)->allocation.width;
-			item->_priv->preferred_height=GTK_WIDGET (item)->allocation.height;
-		}
+        gboolean isFloating = FALSE;
+        gint width=0, height=0, x=0, y = 0;
+        
+        if (GDL_IS_DOCK (gdl_dock_object_get_parent_object (GDL_DOCK_OBJECT (item))))
+        {
+            GdlDock* dock = GDL_DOCK (gdl_dock_object_get_parent_object (GDL_DOCK_OBJECT (item)));
+            g_object_get (dock,
+                          "floating", &isFloating, 
+                          "width", &width,
+                          "height",&height,
+                          "floatx",&x,
+                          "floaty",&y,
+                          NULL);
+        } else {
+            item->_priv->preferred_width=GTK_WIDGET (item)->allocation.width;
+            item->_priv->preferred_height=GTK_WIDGET (item)->allocation.height;
+        }
         item->_priv->ph = GDL_DOCK_PLACEHOLDER (
             g_object_new (GDL_TYPE_DOCK_PLACEHOLDER,
                           "sticky", FALSE,
                           "host", item,
                           "width", width,
                           "height", height,
-						  "floating", isFloating,
-						  "floatx", x,
-						  "floaty", y,
-						  NULL));
+                          "floating", isFloating,
+                          "floatx", x,
+                          "floaty", y,
+                          NULL));
         g_object_ref (item->_priv->ph);
         gtk_object_sink (GTK_OBJECT (item->_priv->ph));
     }
@@ -1693,48 +1781,48 @@ gdl_dock_item_show_item (GdlDockItem *item)
     g_return_if_fail (item != NULL);
 
     GDL_DOCK_OBJECT_UNSET_FLAGS (item, GDL_DOCK_ICONIFIED);
-			
+    
     if (item->_priv->ph) {
-		gboolean isFloating=FALSE;
-		gint width = 0, height = 0, x= 0, y = 0;
-		g_object_get (G_OBJECT(item->_priv->ph),
-					  "width", &width,
-					  "height", &height,
-					  "floating",&isFloating,
-					  "floatx", &x,
-					  "floaty", &y,
-					  NULL);
-		if (isFloating) {
-			GdlDockObject *controller = gdl_dock_master_get_controller (GDL_DOCK_OBJECT_GET_MASTER (item));
-			gdl_dock_add_floating_item (GDL_DOCK (controller),
-										item, x, y, width, height);
-        }else{
-			gtk_container_add (GTK_CONTAINER (item->_priv->ph), GTK_WIDGET (item));
-		}
+        gboolean isFloating=FALSE;
+        gint width = 0, height = 0, x= 0, y = 0;
+        g_object_get (G_OBJECT(item->_priv->ph),
+                      "width", &width,
+                      "height", &height,
+                      "floating",&isFloating,
+                      "floatx", &x,
+                      "floaty", &y,
+                      NULL);
+        if (isFloating) {
+            GdlDockObject *controller =
+                gdl_dock_master_get_controller (GDL_DOCK_OBJECT_GET_MASTER (item));
+            gdl_dock_add_floating_item (GDL_DOCK (controller),
+                                        item, x, y, width, height);
+        } else {
+            gtk_container_add (GTK_CONTAINER (item->_priv->ph),
+                               GTK_WIDGET (item));
+        }
         g_object_unref (item->_priv->ph);
         item->_priv->ph = NULL;
-    }
-    else if (gdl_dock_object_is_bound (GDL_DOCK_OBJECT (item))) {
+        
+    } else if (gdl_dock_object_is_bound (GDL_DOCK_OBJECT (item))) {
         GdlDockObject *toplevel;
-
-	toplevel = gdl_dock_master_get_controller (
-	            GDL_DOCK_OBJECT_GET_MASTER (item));
- 
-	if (item->behavior & GDL_DOCK_ITEM_BEH_NEVER_FLOATING) {
-	    g_warning("Object %s has no default position and flag GDL_DOCK_ITEM_BEH_NEVER_FLOATING is set.\n",
-			GDL_DOCK_OBJECT(item)->name);
-	}
-	else if(toplevel) {
+        
+        toplevel = gdl_dock_master_get_controller
+                        (GDL_DOCK_OBJECT_GET_MASTER (item));
+        
+        if (item->behavior & GDL_DOCK_ITEM_BEH_NEVER_FLOATING) {
+            g_warning("Object %s has no default position and flag GDL_DOCK_ITEM_BEH_NEVER_FLOATING is set.\n",
+                      GDL_DOCK_OBJECT(item)->name);
+        } else if (toplevel) {
             gdl_dock_object_dock (toplevel, GDL_DOCK_OBJECT (item),
                                   GDL_DOCK_FLOATING, NULL);
-        }
-	else
-	    g_warning("There is no toplevel window. GdlDockItem %s cannot be shown.\n", GDL_DOCK_OBJECT(item)->name);
-
-    }
-    else
-	g_warning("GdlDockItem %s is not bound. It cannot be shown.\n", GDL_DOCK_OBJECT(item)->name);
-
+        } else
+            g_warning("There is no toplevel window. GdlDockItem %s cannot be shown.\n", GDL_DOCK_OBJECT(item)->name);
+        
+    } else
+        g_warning("GdlDockItem %s is not bound. It cannot be shown.\n",
+                  GDL_DOCK_OBJECT(item)->name);
+    
     gtk_widget_show (GTK_WIDGET (item));
 }
 
@@ -1766,8 +1854,7 @@ gdl_dock_item_set_default_position (GdlDockItem   *item,
             g_object_ref (reference);
             gtk_object_sink (GTK_OBJECT (reference));
             item->_priv->ph = GDL_DOCK_PLACEHOLDER (reference);
-        }
-        else {
+        } else {
             item->_priv->ph = GDL_DOCK_PLACEHOLDER (
                 g_object_new (GDL_TYPE_DOCK_PLACEHOLDER,
                               "sticky", TRUE,
